from typing import Any

from functional_parsing_library.parser import Parser, T


def ignore_right(left: Parser[T], right: Parser[Any]) -> Parser[T]:
    """
    Parses using the left parser, and then parses using the right. The result from the left parser is discarded. If the
    right parser is unable to match something, the parser fails. The `<` operator is overloaded to call this function.
    For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = char('a') < char('b')
    >>> parser('ab').result
    'a'
    >>> from functional_parsing_library.parser import CouldNotParse
    >>> isinstance(parser('a'), CouldNotParse)
    True
    """
    return (lambda t, s: t) * left & right



from typing import Any

from functional_parsing_library.parser import Parser, S


def ignore_left(left: Parser[Any], right: Parser[S]) -> Parser[S]:
    """
    Parses using the left parser, discards the result, and then parses using the right. The `>` operator is overloaded
    to call this function. For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = char('a') > char('b')
    >>> parser('ab').result
    'b'
    """
    return (lambda t, s: s) * left & right



from functional_parsing_library.strings.modules.ascii import ascii_letter, ascii_lowercase, ascii_uppercase
from functional_parsing_library.strings.modules.char import char
from functional_parsing_library.strings.modules.char_satisfies import char_satisfies
from functional_parsing_library.strings.modules.word import word
from functional_parsing_library.strings.modules.char_in import char_in
from functional_parsing_library.strings.modules.any_char import any_char
from functional_parsing_library.strings.modules.char_does_not_match import char_does_not_match
from functional_parsing_library.strings.modules.char_not_in import char_not_in
from functional_parsing_library.strings.modules.digit import digit

__all__ = [
    'any_char',
    'char',
    'char_does_not_match',
    'char_in',
    'char_not_in',
    'char_satisfies',
    'digit',
    'word',
    'ascii_letter',
    'ascii_lowercase',
    'ascii_uppercase',
]



from collections.abc import Container

from functional_parsing_library.parser import Parser
from functional_parsing_library.strings.modules.char_satisfies import char_satisfies


def char_not_in(characters: Container[str]) -> Parser[str]:
    """
    Match any character not among the given characters. For example,
    >>> char_not_in('abcd')('eddy').result
    'e'
    """
    return char_satisfies(lambda c: c not in characters)



from inspect import signature
from typing import Callable, TypeVarTuple, overload, TypeVar, TypeGuard, Any

from functional_parsing_library.parser import Parser, T, S, ParseResults, CouldNotParse, MappedParser

Ts = TypeVarTuple('Ts')
U = TypeVar('U')


def to_int(string: str) -> int:
    return int(string)


def _number_of_arguments(function: Callable[..., Any]) -> int:
    if function in (int, str, float, dict, set, list, tuple, bool):
        return 1
    return len(signature(function).parameters)


def _accepts_single_argument(
    function: Callable[[T, U, *Ts], S] | Callable[[T], S],
) -> TypeGuard[Callable[[T], S]]:
    return _number_of_arguments(function) == 1


def _accepts_many_arguments(
    function: Callable[[T, U, *Ts], S] | Callable[[T], S],
) -> TypeGuard[Callable[[T, U, *Ts], S]]:
    return not _accepts_single_argument(function)


@overload
def fmap(function: Callable[[T], S], parser: Parser[T]) -> Parser[S]:
    pass


@overload
def fmap(function: Callable[[T, U, *Ts], S], parser: Parser[T]) -> MappedParser[S, U, *Ts]:
    pass


def fmap(
    function: Callable[[T], S] | Callable[[T, U, *Ts], S],
    parser: Parser[T],
) -> Parser[S] | MappedParser[S, U, *Ts]:
    """
    Used to map over parsers. The `*` operator is overloaded to call this function. Given a parser `p` of type
    `Parser[T]` and a callable `f` of type `Callable[[T], S]`, the parser `f * p` will parse using `p`, and then
    apply `f` to the result, resulting in an object of type S. For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = (lambda x: x + 'b') * char('a')
    >>> parser('a').result
    'ab'
    """
    if _accepts_single_argument(function):
        return _fmap_for_one_argument(function, parser)

    if _accepts_many_arguments(function):
        return _fmap_for_multiple_arguments(function, parser)

    raise Exception('Function should accept either one or many arguments')


def _fmap_for_one_argument(
    function: Callable[[T], S],
    parser: Parser[T],
) -> Parser[S]:
    def mapped_parser(to_parse: str) -> ParseResults[S] | CouldNotParse:
        result = parser(to_parse)
        if isinstance(result, CouldNotParse):
            return result
        return ParseResults(
            result=function(result.result),
            remainder=result.remainder
        )

    return Parser(mapped_parser)


def _fmap_for_multiple_arguments(
    function: Callable[[T, U, *Ts], S],
    parser: Parser[T],
) -> MappedParser[S, U, *Ts]:
    def mapped_parser(to_parse: str) -> ParseResults[Callable[[U, *Ts], S]] | CouldNotParse:
        result = parser(to_parse)
        if isinstance(result, CouldNotParse):
            return result
        return ParseResults(
            result=lambda u, *ts: function(result.result, u, *ts),
            remainder=result.remainder
        )

    p: MappedParser[S, U, *Ts] = MappedParser(mapped_parser)
    if _number_of_arguments(function) > 2:
        p.is_multi_arg = True

    return p



from dataclasses import dataclass
from typing import Any

from functional_parsing_library.parser import Parser, CouldNotParse, ParseResults


@dataclass(frozen=True)
class TrySucceeded:
    pass


def try_parser(parser: Parser[Any]) -> Parser[TrySucceeded]:
    """
    try_parser(parser) succeeds when parser does, but consumes nothing
    >>> from functional_parsing_library.combinators import try_parser
    >>> from functional_parsing_library.strings import char
    >>> try_parser(char('a'))('a')
    ParseResults(result=TrySucceeded(), remainder='a')
    """
    def parser_function(to_parse: str) -> ParseResults[TrySucceeded] | CouldNotParse:
        parse_result = parser(to_parse)
        if isinstance(parse_result, CouldNotParse):
            return parse_result
        return ParseResults(TrySucceeded(), to_parse)

    return Parser(parser_function)



from functional_parsing_library.parser import Parser, U, ParseResults, CouldNotParse


def maybe(parser: Parser[U]) -> Parser[U | None]:
    """
    Transforms a failure to parse by the wrapped parser to a successful parse, with None as its result. For example,
    >>> from functional_parsing_library.strings import char
    >>> maybe(char('a'))('b').result is None
    True
    """
    def maybe_parser(to_parse: str) -> ParseResults[U | None]:
        result = parser(to_parse)
        if isinstance(result, CouldNotParse):
            return ParseResults(None, to_parse)
        return result

    return Parser(maybe_parser)



import string

from functional_parsing_library.fmap import to_int
from functional_parsing_library.strings.modules.char import char
from functional_parsing_library.strings.modules.char_in import char_in
from functional_parsing_library.combinators.sequence.many import many


digit = char_in(string.digits)
digits = ''.join * many(digit)
nonnegative_integer = to_int * digits


def negate(x: int) -> int:
    return -x


negative_integer = negate * (char('-') > nonnegative_integer)
integer = nonnegative_integer | negative_integer



from typing import Any

from functional_parsing_library.combinators.sequence.many import some
from functional_parsing_library.parser import Parser, T
from functional_parsing_library.strings.modules.word import word


def separated_by(parser: Parser[T], separator: Parser[Any]) -> Parser[list[T]]:
    """
    Matches one or more (and as many as possible) matches of parser, separated by matches for separator.
    For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = separated_by(char('a'), char(','))
    >>> parser('a,a,a').result
    ['a', 'a', 'a']
    """
    return (lambda t, ts: [t, *ts]) * parser & some(separator > parser)


nothing = word('')


def some_separated_by(parser: Parser[T], separator: Parser[Any]) -> Parser[list[T]]:
    """
    Matches zero or more (and as many as possible) matches of parser, separated by matches for separator.
    For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = some_separated_by(char('a'), char(','))
    >>> parser('a,a,a').result
    ['a', 'a', 'a']
    >>> parser('shibby').result
    []
    """
    return separated_by(parser, separator) | ((lambda x: []) * nothing)



from functional_parsing_library.parser import Parser, T, ParseResults, CouldNotParse


def some(parser: Parser[T]) -> Parser[list[T]]:
    """
    Parse at zero or more, and as many as possible, matches for parser. For example,
    >>> from functional_parsing_library.strings import char
    >>> some(char('a'))('aaab').result
    ['a', 'a', 'a']
    >>> some(char('a'))('b').result
    []
    """
    def parser_(to_parse: str) -> ParseResults[list[T]]:
        remainder = to_parse
        result: list[T] = []
        while remainder:
            one_parse = parser(remainder)
            if isinstance(one_parse, CouldNotParse):
                break
            result.append(one_parse.result)
            remainder = one_parse.remainder

        return ParseResults(result=result, remainder=remainder)

    return Parser(parser_)


def many(parser: Parser[T]) -> Parser[list[T]]:
    """
    Parse at least one, and as many as possible, matches for parser. For example,
    >>> from functional_parsing_library.strings import char
    >>> many(char('a'))('aaab').result
    ['a', 'a', 'a']
    """
    return (lambda t, ts: [t, *ts]) * parser & some(parser)



from __future__ import annotations

from typing import Generic

from functional_parsing_library.parser import Parser, T, ParseResults, CouldNotParse


class RecursiveParser(Generic[T]):
    """
    A container for forward references to parsers so as to support parsers for recursive languages, such as the language
    of expressions of the form (), (3), ((3)), 3, ((())).
    """
    def __init__(self) -> None:
        self._parser: Parser[T] | None = None

    @property
    def parser(self) -> Parser[T]:
        return Parser(self._parse_function)

    @parser.setter
    def parser(self, parser: Parser[T]) -> None:
        self._parser = parser

    def _parse_function(self, to_parse: str) -> ParseResults[T] | CouldNotParse:
        if self._parser is None:
            raise NotImplementedError
        return self._parser(to_parse)



from dataclasses import dataclass

from functional_parsing_library.asserts import assert_parsing_succeeds
from functional_parsing_library.combinators.sequence.many import many
from functional_parsing_library.combinators.sequence.separated_by import separated_by
from functional_parsing_library.integer.integer import integer
from functional_parsing_library.strings.modules.char import char
from functional_parsing_library.strings.modules.char_not_in import char_not_in

Value = str | int


@dataclass(frozen=True)
class CSVFile:
    header: list[str]
    records: list[list[Value]]


comma = char(',')
string_char = char_not_in(',\n')
column_name = ''.join * many(string_char)
new_line = char('\n')
header = separated_by(column_name, separator=comma) < new_line


string = ''.join * many(string_char)
value = integer | string
record = separated_by(value, separator=comma)
records = separated_by(record, separator=new_line)

csv_file = CSVFile * header & records


def test_that_stuff_parses() -> None:
    my_file = '\n'.join([
        'hi,mom',
        'orf,blorf',
        '3,dorf',
        '9,nein9'
    ])

    assert_parsing_succeeds(csv_file, my_file).with_result(
        CSVFile(
            header=['hi', 'mom'],
            records=[
                ['orf', 'blorf'],
                [3, 'dorf'],
                [9, 'nein9'],
            ]
        )
    )



from functional_parsing_library.parser import Parser
from functional_parsing_library.strings.modules.char_satisfies import char_satisfies


def char(c: str) -> Parser[str]:
    """
    Creates a parser which parses a single character c. For example,
    >>> char('a')('aabcd').result
    'a'
    >>> char('a')('aabcd').remainder
    'abcd'
    >>> from functional_parsing_library.parser import CouldNotParse
    >>> isinstance(char('a')('bla'), CouldNotParse)
    True
    """

    if len(c) != 1:
        raise ValueError

    return char_satisfies(
        condition=lambda x: x == c,
        reason_factory=lambda to_parse: f'String "{to_parse}" does not start with "{c}"',
    )



from typing import TypeVarTuple, Callable, overload, TypeVar, cast

from functional_parsing_library.parser import Parser, S, T, ParseResults, CouldNotParse, MappedParser

Ts = TypeVarTuple('Ts')


U = TypeVar('U')


@overload
def parse_and(left: MappedParser[S, T], right: Parser[T]) -> Parser[S]:
    pass


@overload
def parse_and(left: MappedParser[S, T, U, *Ts], right: Parser[T]) -> MappedParser[S, U, *Ts]:
    pass


def parse_and(
    left: MappedParser[S, T] | MappedParser[S, T, U, *Ts],
    right: Parser[T],
) -> Parser[S] | MappedParser[S, U, *Ts]:
    """
    Is used to combine parsers p and q to a parser which first parses using p and then the remainder using q. The `&`
    operator is overloaded to make use of this function.

    If p is of type Parser[T] and q is of type Parser[S], then it is not obvious what type p & q should have. For example,
    it could be of type Parser[tuple[T, S]], it could be of type Parser[list[T | S]], and so on. Therefore, `&` can only
    be used in combination with a callable f of type `Callable[[T, S], U]` which specifies how to combine the result of
    p with the result of q. Writing `p & q` without applying such a function will raise a TypeError.

    For example,
    >>> from functional_parsing_library.strings import char
    >>> a, b = char('a'), char('b')
    >>> plus = lambda x, y: x + ' and ' + y
    >>> parser = plus * a & b
    >>> parser('ab').result
    'a and b'

    This works with callables of more than two arguments as well:
    >>> c = char('c')
    >>> triple_plus = lambda x, y, z: x + y + z
    >>> parser = triple_plus * a & b & c
    >>> parser('abc').result
    'abc'
    """
    if left.is_multi_arg:
        multi_arg_left = cast(MappedParser[S, T, U, *Ts], left)
        return _parse_and_for_multiple_arguments(multi_arg_left, right)

    single_arg_left = cast(MappedParser[S, T], left)
    return _parse_and_for_single_argument(single_arg_left, right)


def _parse_and_for_multiple_arguments(
    left: MappedParser[S, T, U, *Ts],
    right: Parser[T],
) -> MappedParser[S, U, *Ts]:
    def parser_(to_parse: str) -> ParseResults[Callable[[U, *Ts], S]] | CouldNotParse:
        result = _parse_left_and_then_right(left, right, to_parse)
        if isinstance(result, CouldNotParse):
            return result
        right_result, parsed_function = result

        return ParseResults(
            result=lambda *ts: parsed_function(right_result.result, *ts),
            remainder=right_result.remainder
        )

    return MappedParser(parser_)


def _parse_and_for_single_argument(
    left: MappedParser[S, T],
    right: Parser[T],
) -> Parser[S]:
    def parser(to_parse: str) -> ParseResults[S] | CouldNotParse:
        result = _parse_left_and_then_right(left, right, to_parse)
        if isinstance(result, CouldNotParse):
            return result
        right_result, parsed_function = result

        return ParseResults(
            result=parsed_function(right_result.result),
            remainder=right_result.remainder
        )

    return Parser(parser)


def _parse_left_and_then_right(
    left: MappedParser[S, T, *Ts],
    right: Parser[T],
    to_parse: str
) -> tuple[ParseResults[T], Callable[[T, *Ts], S]] | CouldNotParse:
    left_result = left(to_parse)
    if isinstance(left_result, CouldNotParse):
        return left_result

    right_result = right(left_result.remainder)
    if isinstance(right_result, CouldNotParse):
        return right_result

    return right_result, left_result.result



from .sequence.many import many, some
from functional_parsing_library.combinators.sequence.until.many_till_exclusive import many_till_exclusive
from .sequence.separated_by import separated_by, some_separated_by
from .sequence.until.many_till import many_till
from .sequence.until.some_till import some_till
from .sequence.until.some_till_exclusive import some_till_exclusive
from .try_parser_ import try_parser

__all__ = [
    'many',
    'some',
    'many_till',
    'many_till_exclusive',
    'separated_by',
    'some_separated_by',
    'some_till',
    'some_till_exclusive',
    'try_parser',
]



from typing import Any

from functional_parsing_library.parser import Parser, U, ParseResults, CouldNotParse


def some_till_exclusive(parser: Parser[U], until: Parser[Any]) -> Parser[list[U]]:
    """
    Parses zero or more matches for parser, followed by a match for until. Does not consume the match for until.
    For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = some_till_exclusive(char('a'), char('b'))
    >>> parser('aab').result
    ['a', 'a']
    >>> parser('aab').remainder
    'b'
    >>> parser('b').result
    []
    """
    def parse_function(to_parse: str) -> ParseResults[list[U]] | CouldNotParse:
        results: list[U] = []
        remainder = to_parse
        while True:
            delimiter = until(remainder)
            if not isinstance(delimiter, CouldNotParse):
                return ParseResults(results, remainder)
            parse_results = parser(remainder)
            if isinstance(parse_results, CouldNotParse):
                return parse_results
            results.append(parse_results.result)
            remainder = parse_results.remainder

    return Parser(parse_function)



from typing import Any

from functional_parsing_library.combinators import try_parser
from functional_parsing_library.parser import Parser, S


def lookahead(parser: Parser[S], look_for: Parser[Any]) -> Parser[S]:
    """
    Matches the same as the first parser, as long as it is followed by something matching the second parser. Does not
    consume whatever the second parser matches. For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = lookahead(char('a'), char('b'))
    >>> parser('abbb').result
    'a'
    >>> parser('abbb').remainder
    'bbb'
    >>> from functional_parsing_library.parser import CouldNotParse
    >>> isinstance(parser('a'), CouldNotParse)
    True
    """
    return (lambda t, _: t) * parser & try_parser(look_for)



import string
from typing import Any

from functional_parsing_library.combinators.sequence.many import many
from functional_parsing_library.combinators.sequence.separated_by import some_separated_by
from functional_parsing_library.integer.integer import integer
from functional_parsing_library.recursive_parser import RecursiveParser
from functional_parsing_library.strings.modules.char import char
from functional_parsing_library.strings.modules.char_in import char_in
from functional_parsing_library.strings.modules.word import word


valid_json_string_chars = string.ascii_letters + 'idk'
json_string = ''.join * ((char('"') > many(char_in(valid_json_string_chars))) < char('"'))

_json: RecursiveParser[Any] = RecursiveParser()
array = (char('[') > some_separated_by(_json.parser, word(', '))) < char(']')
key_value_pair = (lambda t, s: (t, s)) * (json_string < word(': ')) & _json.parser
json_object = dict * (
    (char('{') > some_separated_by(key_value_pair, word(', '))) < char('}')
)
_json.parser = integer | array | json_string | json_object
json = _json.parser



from __future__ import annotations

from typing import Any, Generic, Self

from functional_parsing_library.parser import Parser, S, CouldNotParse, ParseResults, FailureReason


class _ParsingSuccessTestResult(Generic[S]):
    def __init__(
        self,
        actual: ParseResults[S],
    ) -> None:
        self._actual = actual

    def with_result(self, expected: S) -> _ParsingSuccessTestResult[S]:
        assert self._actual.result == expected
        return self

    def with_remainder(self, expected: str) -> _ParsingSuccessTestResult[S]:
        assert self._actual.remainder == expected
        return self


def assert_parsing_succeeds(parser: Parser[S], to_parse: str) -> _ParsingSuccessTestResult[S]:
    result = parser(to_parse)
    assert not isinstance(result, CouldNotParse)
    return _ParsingSuccessTestResult(
        actual=result,
    )


class _ParsingFailureTestResult:
    def __init__(self, result: CouldNotParse) -> None:
        self._result = result

    def with_reason(self, reason: FailureReason) -> Self:
        assert self._result.reason == reason
        return self


def assert_parsing_fails(parser: Parser[Any], to_parse: str) -> _ParsingFailureTestResult:
    result = parser(to_parse)
    assert isinstance(result, CouldNotParse)
    return _ParsingFailureTestResult(result)



from __future__ import annotations

from dataclasses import dataclass

from functional_parsing_library.asserts import assert_parsing_succeeds
from functional_parsing_library.integer.integer import integer
from functional_parsing_library.recursive_parser import RecursiveParser
from functional_parsing_library.strings import char


@dataclass(frozen=True)
class Plus:
    left_operand: Expression
    right_operand: Expression


Leaf = int
Expression = Leaf | Plus

leaf = integer
plus: RecursiveParser[Plus] = RecursiveParser()
operand = ((char('(') > plus.parser) < char(')')) | leaf
plus.parser = Plus * (operand < char('+')) & operand

expression = plus.parser | leaf


def test_computation() -> None:
    to_parse = '300+(200+900)'
    assert_parsing_succeeds(expression, to_parse).with_result(
        Plus(
            left_operand=300,
            right_operand=Plus(
                left_operand=200,
                right_operand=900,
            )
        )
    )



from typing import Any

from functional_parsing_library.combinators.sequence.until.some_till_exclusive import some_till_exclusive
from functional_parsing_library.parser import Parser, U


def many_till_exclusive(parser: Parser[U], until: Parser[Any]) -> Parser[list[U]]:
    """
    Parses one or more matches for parser, followed by a match for until. Does not consume the match for until.
    For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = many_till_exclusive(char('a'), char('b'))
    >>> parser('aab').result
    ['a', 'a']
    >>> parser('aab').remainder
    'b'
    """
    return (lambda t, ts: [t, *ts]) * parser & some_till_exclusive(parser, until)



from functional_parsing_library.parser import Parser, T, S, ParseResults, CouldNotParse


def or_2(parser_1: Parser[T], parser_2: Parser[S]) -> Parser[T | S]:
    """
    Combines two parsers into one which tries to match for the first one, and upon failure match for the second one.
    The `|` operator is overloaded to call this function. For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = char('a') | char('b')
    >>> parser('a').result
    'a'
    >>> parser('b').result
    'b'
    """
    def parser(to_parse: str) -> ParseResults[T | S] | CouldNotParse:
        try_1 = parser_1(to_parse)
        if not isinstance(try_1, CouldNotParse):
            return try_1
        return parser_2(to_parse)

    return Parser(parser)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar, Generic, Callable, overload, TypeVarTuple

T = TypeVar('T', covariant=True)
Ts = TypeVarTuple('Ts')
S = TypeVar('S')
U = TypeVar('U')


@dataclass(frozen=True)
class ParseResults(Generic[T]):
    result: T
    remainder: str


@dataclass(frozen=True)
class NoReasonSpecified:
    pass


FailureReason = NoReasonSpecified | str


@dataclass(frozen=True)
class CouldNotParse:
    reason: FailureReason = NoReasonSpecified()


class Parser(Generic[T]):
    """
    This class is a wrapper for functions of signature `str -> ParseResults[T] | CouldNotParse` (i.e., parsers), and
    is used to overload the operators *, |, &, <, and > to make them call various parser combinators. If you wish to
    implement a parser which cannot be implemented by combining parsers found in this library, you implement a function
    `f` of signature `str -> ParseResults[T] | CouldNotParse`, and wrap it like so: `parser = Parser(f)`. Now your
    parser is ready for use and combination with other parsers.
    """
    def __init__(self, parser_function: Callable[[str], ParseResults[T] | CouldNotParse]) -> None:
        self._parser_function = parser_function

    def __call__(self, to_parse: str) -> ParseResults[T] | CouldNotParse:
        return self._parser_function(to_parse)

    def __or__(self, other: Parser[S]) -> Parser[T | S]:
        from functional_parsing_library.combinators.parse_or import or_2
        return or_2(parser_1=self, parser_2=other)

    @overload
    def __rmul__(self, other: Callable[[T, U, *Ts], S]) -> MappedParser[S, U, *Ts]:
        pass

    @overload
    def __rmul__(self, other: Callable[[T], S]) -> Parser[S]:
        pass

    def __rmul__(self, other: Callable[[T, U, *Ts], S] | Callable[[T], S]) -> MappedParser[S, U, *Ts] | Parser[S]:
        from functional_parsing_library.fmap import fmap
        return fmap(function=other, parser=self)

    def __gt__(self, other: Parser[S]) -> Parser[S]:
        from functional_parsing_library.combinators.ignore.ignore_left import ignore_left
        return ignore_left(left=self, right=other)

    def __lt__(self, other: Parser[S]) -> Parser[T]:
        from functional_parsing_library.combinators.ignore.ignore_right import ignore_right
        return ignore_right(left=self, right=other)

    @overload
    def __rand__(self, other: MappedParser[S, T]) -> Parser[S]:
        pass

    @overload
    def __rand__(self, other: MappedParser[S, T, U, *Ts]) -> MappedParser[S, U, *Ts]:
        pass

    def __rand__(self, other: MappedParser[S, T] | MappedParser[S, T, U, *Ts]) -> Parser[S] | MappedParser[S, U, *Ts]:
        from functional_parsing_library.combinators.parse_and import parse_and
        return parse_and(left=other, right=self)


class MappedParser(Parser[Callable[[*Ts], S]], Generic[S, *Ts]):
    """
    This class exists only to make __rand__ work. In the expression
    a & b, only if a and b are of different types is __rand__ evaluated.
    If a and b are of the same type, __rand__ is implemented, and __and__ is
    not, an error is raised.
    """
    def __init__(
        self,
        parser_function: Callable[[str], ParseResults[Callable[[*Ts], S]] | CouldNotParse],
    ) -> None:
        super().__init__(parser_function)
        self.is_multi_arg = False



from typing import Any

from functional_parsing_library.combinators.sequence.until.many_till_exclusive import many_till_exclusive
from functional_parsing_library.parser import Parser, U


def many_till(parser: Parser[U], until: Parser[Any]) -> Parser[list[U]]:
    """
    Parses one or more matches for parser, followed by a match for until. Consumes the match for until.
    For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = many_till(char('a'), char('b'))
    >>> parser('aab').result
    ['a', 'a']
    >>> parser('aab').remainder
    ''
    """
    return many_till_exclusive(parser, until) < until



from typing import Any

from functional_parsing_library.check_for_empty_string import check_for_empty_string
from functional_parsing_library.parser import Parser, ParseResults, CouldNotParse
from functional_parsing_library.strings.modules.pop_one_character import pop_one_character


def char_does_not_match(parser: Parser[Any]) -> Parser[str]:
    """
    Match any character not matching the given parser. For example,
    >>> from functional_parsing_library.strings import digit
    >>> non_digit = char_does_not_match(digit)
    >>> non_digit('a').result
    'a'
    >>> isinstance(non_digit('3'), CouldNotParse)
    True
    """
    @check_for_empty_string
    def parser_(to_parse: str) -> ParseResults[str] | CouldNotParse:
        result = parser(to_parse)
        if not isinstance(result, CouldNotParse):
            return CouldNotParse()

        return pop_one_character(to_parse)

    return Parser(parser_)



from typing import Any

from functional_parsing_library.combinators.sequence.until.some_till_exclusive import some_till_exclusive
from functional_parsing_library.parser import Parser, U


def some_till(parser: Parser[U], until: Parser[Any]) -> Parser[list[U]]:
    """
    Parses zero or more matches for parser, followed by a match for until. Consumes the match for until. For example,
    >>> from functional_parsing_library.strings import char
    >>> parser = some_till(char('a'), char('b'))
    >>> parser('aab').result
    ['a', 'a']
    >>> parser('aab').remainder
    ''
    >>> parser('b').result
    []
    """
    return some_till_exclusive(parser, until) < until



from functional_parsing_library.parser import Parser, ParseResults, CouldNotParse


def word(word_to_parse_for: str) -> Parser[str]:
    """
    Parse a given word. For example,
    >>> parser = word('train')
    >>> parser('training').result
    'train'
    >>> parser('training').remainder
    'ing'
    """
    def parser(to_parse: str) -> ParseResults[str] | CouldNotParse:
        if not to_parse.startswith(word_to_parse_for):
            return CouldNotParse(f'String "{to_parse}" does not start with "{word_to_parse_for}"')
        return ParseResults(word_to_parse_for, to_parse[len(word_to_parse_for):])

    return Parser(parser)



import string
from enum import auto, Enum
from typing import Callable, Any

from functional_parsing_library.combinators.sequence.many import many
from functional_parsing_library.parser import S
from functional_parsing_library.strings.modules.char import char
from functional_parsing_library.strings.modules.char_in import char_in
from functional_parsing_library.strings.modules.word import word


class Honorific(Enum):
    MR = auto()
    MS = auto()


class Person:
    def __init__(self, honorific: Honorific, name: str, surname: str) -> None:
        self._honorific = honorific
        self._name = name
        self._surname = surname


def const(value: S) -> Callable[[Any], S]:
    return lambda x: value


alphabetic_character = char_in(string.ascii_letters)
alphabetic_word = ''.join * many(alphabetic_character)
mr = const(Honorific.MR) * (word('Mr.') | word('mr.'))
ms = const(Honorific.MS) * (word('Ms.') | word('ms.'))
honorific = (mr | ms) < char(' ')
name = alphabetic_word < char(' ')
surname = alphabetic_word
person = Person * honorific & name & surname



from string import ascii_lowercase as ascii_lowercase_letters, ascii_uppercase as ascii_uppercase_letters
from functional_parsing_library.strings.modules.char_in import char_in

ascii_lowercase = char_in(ascii_lowercase_letters)

ascii_uppercase = char_in(ascii_uppercase_letters)

ascii_letter = ascii_lowercase | ascii_uppercase



from typing import Callable

from functional_parsing_library.check_for_empty_string import check_for_empty_string
from functional_parsing_library.parser import Parser, ParseResults, CouldNotParse, FailureReason, NoReasonSpecified
from functional_parsing_library.strings.modules.pop_one_character import pop_one_character


def char_satisfies(
    condition: Callable[[str], bool],
    reason_factory: Callable[[str], FailureReason] = lambda to_parse: NoReasonSpecified(),
) -> Parser[str]:
    """
    Match any character satisfying the given condition. For example,
    >>> char_satisfies(lambda c: c == 'b')('b').result
    'b'
    """
    @check_for_empty_string
    def parser(to_parse: str) -> ParseResults[str] | CouldNotParse:
        if condition(to_parse[0]):
            return pop_one_character(to_parse)
        return CouldNotParse(reason_factory(to_parse))
    return Parser(parser)



from functional_parsing_library.parser import ParseResults


def pop_one_character(string: str) -> ParseResults[str]:
    return ParseResults(
        result=string[0],
        remainder=string[1:],
    )



from functional_parsing_library.parser import Parser, S, CouldNotParse


class ParsingError(Exception):
    pass


class NotFullyParsed(Exception):
    pass


def run_parser(
    parser: Parser[S],
    to_parse: str,
    raise_on_remainder: bool = False,
) -> S:
    """
    Utility function for running parsers. On parse errors, instead of an error value CouldNotParse, this function raises
    an exception. It also unwraps the parse result:
    >>> from functional_parsing_library.strings import char
    >>> parser = char('a')
    >>> run_parser(parser, 'ab')
    'a'
    """
    parse_result = parser(to_parse)

    if isinstance(parse_result, CouldNotParse):
        raise ParsingError

    if raise_on_remainder and len(parse_result.remainder) > 0:
        raise NotFullyParsed

    return parse_result.result



from collections.abc import Container

from functional_parsing_library.parser import Parser
from functional_parsing_library.strings.modules.char_satisfies import char_satisfies


def char_in(string: Container[str]) -> Parser[str]:
    """
    Matches any character in the given string. For example,
    >>> parser = char_in('abcd')
    >>> parser('abdicate').result
    'a'
    >>> parser('abdicate').remainder
    'bdicate'
    """
    return Parser(char_satisfies(lambda c: c in string))



from dataclasses import dataclass

from functional_parsing_library.asserts import assert_parsing_succeeds, assert_parsing_fails
from functional_parsing_library.integer.integer import nonnegative_integer
from functional_parsing_library.strings import char


@dataclass(frozen=True, order=True)
class Version:
    major: int
    minor: int
    patch: int


period = char('.')
version = Version * (nonnegative_integer < period) & (nonnegative_integer < period) & nonnegative_integer


def test_version() -> None:
    assert_parsing_succeeds(version, '1.2.0').with_result(Version(1, 2, 0))


def test_version_fails_on_negatives() -> None:
    assert_parsing_fails(version, '1.-2.0')
